#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
from dataclasses import Field
from typing import Any

import sqlalchemy.orm
from dynamic_types.class_name import _prepare_class_name

from .abstract import AbstractView
from .utils import get_related_entity_type, get_remote_column
from ..entity import Entity


class RelationView(AbstractView, abc.ABC):
    @classmethod
    def _produce_view(cls, relation: sqlalchemy.orm.RelationshipProperty) -> type:
        if not isinstance(relation, sqlalchemy.orm.RelationshipProperty):
            raise TypeError(f'RelationInsertView subscription requires RelationshipProperty, got: {relation}')
        related_entity_type: type[Entity] = get_related_entity_type(relation)
        relation_remote_column = get_remote_column(relation)
        field_name_to_field = cls._produce_field_name_to_field(related_entity_type, relation_remote_column)
        class_name = _prepare_class_name([part[0].upper() + part[1:] for part in str(relation).split(".")], cls)
        return cls._produce_child_dataclass(class_name, field_name_to_field)

    @classmethod
    @abc.abstractmethod
    def _produce_field_name_to_field(cls, related_entity_type: type[Entity],
                                     relation_remote_column: sqlalchemy.Column) -> dict[str, Field | Any]:
        pass



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import typing
from dataclasses import dataclass

from entity_read.sql.context import SqlContext
# from entity_read.select.atoms.atoms import NodeType
from .node import Node, SQL
from entity_read.sql.atoms.abstract import Selectable
from entity_read.sql.atoms.column import Column


@dataclass(kw_only=True, frozen=True, slots=True)
class Order(Node):
    def to_sql(self, context: SqlContext, key_to_expression=None) -> SQL:
        pass

    # type = NodeType.order
    value: Selectable
    asc: bool = True
    nulls_last: typing.Optional[bool] = None

    def to_order_sql(self, context, key_to_expression: dict = None) -> str:
        expression_str = self.value.to_sql(context, key_to_expression)
        if not self.asc:
            expression_str += " DESC"
        if self.nulls_last is not None:
            expression_str += " NULLS " + ("LAST" if self.nulls_last else "FIRST")
        return expression_str

    def requires(self) -> list[Column]:
        return self.value.requires()

    def get_name(self) -> str:
        nulls_last = f",nulls_last={self.nulls_last}" if self.nulls_last is not None else ''
        return f"order({self.value.get_name()},asc={self.asc}{nulls_last})"



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
from dataclasses import dataclass
from typing import NamedTuple, Optional

from sqlalchemy.orm import InstrumentedAttribute

from .move import descend, ascend


class ProxyStep(NamedTuple):
    name: str
    local_column_to_remote: dict[str, str]
    alias: str = None
    reverse: bool = False


@dataclass
class AttributeProxy:
    name: str
    attribute: InstrumentedAttribute
    path: list[ProxyStep] = dataclasses.field(default_factory=list)
    alias: Optional[str] = None

    def descend(self, alias: Optional[str] = None) -> 'EntityProxy':
        entity_proxy = descend(self.attribute, alias=alias)
        return entity_proxy.prepend_steps(self.path)

    def ascend(self, alias: Optional[str]) -> 'EntityProxy':
        entity_proxy = ascend(self.attribute, alias=alias)
        return entity_proxy.prepend_steps(self.path)

    def label(self, alias: Optional[str]):
        return AttributeProxy(name=self.name, attribute=self.attribute, path=self.path[:], alias=alias)


class EntityProxy:
    def __init__(self, entity, path: list[ProxyStep], alias: Optional[str] = None):
        self._path = path
        self._entity = entity
        if alias:
            last_step_as_dict = self._path[-1]._asdict()
            last_step_as_dict["alias"] = alias
            self._path[-1] = ProxyStep(**last_step_as_dict)
        # self._alias = alias
        # print("init EntityProxy", self)

    def get_entity(self):
        return self._entity

    def get_path(self) -> list[ProxyStep]:
        return self._path

    def get_alias(self) -> Optional[str]:
        return self._alias

    def prepend_steps(self, steps) -> 'EntityProxy':
        # print("prepend_steps", steps, self._path)
        return EntityProxy(self._entity, steps + self._path)

    def __getattr__(self, item: str):
        result = getattr(self._entity, item)
        if isinstance(result, InstrumentedAttribute):
            return AttributeProxy(item, result, self._path[:])
        return object.__getattribute__(self, item)

    def __str__(self):
        cls = type(self)
        return f'{cls.__name__}(entity={self._entity}, path={self._path})'

    def label(self, alias: Optional[str]):
        return EntityProxy(entity=self._entity, path=self._path[:], alias=alias)
